@@ -3,7 +3,7 @@
  *
  *  Copyright (C) 1994  Scott D. Heavner
  *
- *  $Id: nc_inode.c,v 1.17 1996/10/13 17:14:56 sdh Exp $
+ *  $Id: nc_inode.c,v 2.5 2000/03/11 19:07:58 denis Exp $
  */
 
 #include <ctype.h>
@@ -142,16 +142,18 @@ static void cdump_inode_values(unsigned long nr, struct Generic_Inode *GInode, i
   if (fsc->inode->i_links_count)
     mvwprintw(workspace,0,11,"%3d",GInode->i_links_count);
 
-  if (fsc->inode->i_uid)
+  if (fsc->inode->i_uid) {
     if ((NC_PASS = getpwuid(GInode->i_uid))!=NULL)
       mvwprintw(workspace,0,15,"%-8s",NC_PASS->pw_name);
     else
       mvwprintw(workspace,0,15,"%-8d",GInode->i_uid);
-  if (fsc->inode->i_gid)
+  }
+  if (fsc->inode->i_gid) {
     if ((NC_GROUP = getgrgid(GInode->i_gid))!=NULL)
       mvwprintw(workspace,0,24,"%-8s",NC_GROUP->gr_name);
     else
       mvwprintw(workspace,0,24,"%-8d",GInode->i_gid);
+  }
   if (fsc->inode->i_size)
     mvwprintw(workspace,0,32,"%9ld",GInode->i_size);
   if (fsc->inode->i_mtime)
@@ -457,7 +459,7 @@ int inode_mode() {
   unsigned long a;
   static unsigned char *copy_buffer = NULL;
 
-  display_trailer("F1/H for help.  F2/^O for menu.  Q to quit");
+  display_trailer("F1/? for help.  F2/^O for menu.  Q to quit");
 
   GInode = FS_cmd.read_inode(current_inode);
   cdump_inode_values(current_inode, GInode, (highlight_field|LDE_DUMP_ILABELS));
@@ -519,6 +521,7 @@ int inode_mode() {
 	cwrite_inode(current_inode, GInode, &modified);
 	for (i=0;(i<fsc->N_BLOCKS);i++)
 	  fake_inode_zones[i] = GInode->i_zone[i];
+	fake_inode_zones[INODE_BLKS] = GInode->i_size;
 	return CMD_RECOVERY_MODE;
 	break;
 
