@@ -3,7 +3,7 @@
  *
  *  Copyright (C) 1994  Scott D. Heavner
  *
- *  $Id: nc_lde.c,v 1.25 1996/10/11 00:34:57 sdh Exp $
+ *  $Id: nc_lde.c,v 2.5 2000/03/11 19:07:59 denis Exp $
  */
 
 #include <stdio.h>
@@ -23,18 +23,46 @@
 #include "nc_inode.h"
 #include "nc_block.h"
 #include "keymap.h"
-  
+
+#ifdef HAVE_LIBGPM
+#  include <gpm.h>  
+#  undef  getch           /* Supress warning about redefinition */
+#  define getch Gpm_Getch
+#endif
+
+/* Create globals declared extern in nc_lde.h */
+/* Curses variables */
+int    WHITE_ON_BLUE, WHITE_ON_RED, RED_ON_BLACK;
+WINDOW *header, *workspace, *trailer;
+unsigned long current_inode;
+unsigned long current_block;
+unsigned long fake_inode_zones[INODE_BLKS+1];
+
+
+#ifdef HAVE_LIBGPM
+/* Callback for GPM mouse events */
+static int lde_mouse_handler(Gpm_Event *event, void *clientdata)
+{
+  /* should only come in here on single click */
+
+  if ((event->modifiers)||(event->buttons&(GPM_B_MIDDLE|GPM_B_RIGHT)))
+    return 0;
+
+  return ( ((event->x)<<24) | ((event->y)<<16) | 2048 );
+}
+#endif
 
 /* This will recognize escapes keys as "META" strokes */
 int nc_mgetch(void)
 {
   int c;
 
-  if ( (c=getch()) == ESC )
-    if ( (c=getch()) == ESC )
+  if ( (c=getch()) == LDE_ESC ) {
+    if ( (c=getch()) == LDE_ESC )
       return c;
     else
-      return META(c);
+      return LDE_META(c);
+  }
 
   return c;
 }
@@ -83,16 +111,18 @@ int cread_num(char *coutput, unsigned long *a)
 {
   char cinput[80];
 
-#if TRAILER_SIZE>0
-#define window_available trailer
-#define LINE_NUMBER TRAILER_SIZE-1
-#else
-  win *window_avaliable;
-  window_avaliable = newwin(5,WIN_COL,((VERT-5)/2+HEADER_SIZE),HOFF);
-  werase(window_avaliable);
-  box(window_avaliable,0,0);
-#define LINE_NUMBER 0
-#endif
+  lde_flags.quit_now = 0;
+
+# if TRAILER_SIZE>0
+#   define window_available trailer
+#   define LINE_NUMBER TRAILER_SIZE-1
+# else
+    win *window_avaliable;
+    window_avaliable = newwin(5,WIN_COL,((VERT-5)/2+HEADER_SIZE),HOFF);
+    werase(window_avaliable);
+    box(window_avaliable,0,0);
+#  define LINE_NUMBER 0
+# endif
 
   wmove(window_available,LINE_NUMBER,0);
   wclrtoeol(window_available);
@@ -105,14 +135,16 @@ int cread_num(char *coutput, unsigned long *a)
   wmove(window_available,LINE_NUMBER,0);
   wclrtoeol(window_available);
   wrefresh(window_available);
-#if TRAILER_SIZE<=0
-  delwin(window_available);
-#endif
+# if TRAILER_SIZE<=0
+    delwin(window_available);
+# endif
 
   if ((coutput[0]!=' ')&&(a)) {
     *a = read_num(cinput);
     return strlen(cinput);
   } else {
+    if (lde_flags.quit_now)  /* Control-C aborts */
+      return -1;
     strncpy(coutput, cinput, 80);
     return strlen(cinput);
   }
@@ -139,8 +171,10 @@ void update_header(void)
 #if HEADER_SIZE>0
   int i,j;
 
-  mvwprintw(header,HEADER_SIZE-1,HOFF,"Inode: %10lu (0x%8.8lX) ",current_inode,current_inode);
-  mvwprintw(header,HEADER_SIZE-1,HOFF+32,"Block: %10lu (0x%8.8lX) ",current_block,current_block);
+  mvwprintw(header,HEADER_SIZE-1,HOFF,"Inode: %10lu (0x%8.8lX) ",
+	    current_inode,current_inode);
+  mvwprintw(header,HEADER_SIZE-1,HOFF+32,"Block: %10lu (0x%8.8lX) ",
+	    current_block,current_block);
   for (j=0;j<fsc->N_BLOCKS;j++) {
     if (fake_inode_zones[j]) { 
       mvwaddch(header,HEADER_SIZE-1,HOFF+64+j,'-');
@@ -174,17 +208,9 @@ void clobber_window(WINDOW *win)
 void restore_header(void)
 {
 #if HEADER_SIZE>0
-  int i,j;
   char echo_string[132];
 
-  wattron(header,WHITE_ON_BLUE);
-
-  /* A cute way to RVS video the header window, since erase and clear don't seem to
-   * do it for me. */
-  for (i=0;i<HEADER_SIZE;i++)
-    for (j=0;j<COLS;j++)
-      mvwaddch(header,i,j,' ');
-
+  werase(header);
   sprintf(echo_string,"%s v%s : %s : %s",
 	  program_name,VERSION,text_names[fsc->FS],device_name);
   mvwprintw(header,0,(COLS-strlen(echo_string))/2-1,echo_string);
@@ -195,9 +221,9 @@ void restore_header(void)
 /* Redraw header and trailer, other routine does workspace -- ^L */
 void refresh_ht(void)
 {
-#if TRAILER_SIZE>0
-  redraw_win(trailer);
-#endif
+# if TRAILER_SIZE>0
+    redraw_win(trailer);
+# endif
   restore_header();
 }
 
@@ -223,16 +249,16 @@ void nc_warn(char *fmt, ...)
   log_error(echo_string);
  
   if (!lde_flags.quiet) beep();
-#if TRAILER_SIZE>0
-  wmove(trailer,TRAILER_SIZE-1,0);
-  wclrtoeol(trailer);
-  mvwaddstr(trailer,TRAILER_SIZE-1,(COLS-strlen(echo_string))/2-1,
-	    echo_string);
-  wrefresh(trailer);
-  wmove(trailer,TRAILER_SIZE-1,0);
-  wclrtoeol(trailer);
-  /* Still has to be refreshed by someone */
-#endif
+# if TRAILER_SIZE>0
+    wmove(trailer,TRAILER_SIZE-1,0);
+    wclrtoeol(trailer);
+    mvwaddstr(trailer,TRAILER_SIZE-1,(COLS-strlen(echo_string))/2-1,
+	      echo_string);
+    wrefresh(trailer);
+    wmove(trailer,TRAILER_SIZE-1,0);
+    wclrtoeol(trailer);
+    /* Still has to be refreshed by someone */
+# endif
 }
 
 /* Dump the error log to a window */
@@ -262,7 +288,7 @@ int error_popup(void)
 int do_popup_menu(lde_menu menu[], lde_keymap keys[])
 {
   WINDOW *win, *bigwin;
-  int    c, window_length, window_width, length, width, last_highlight = 0, highlight = 0;
+  int    c, result, window_length, window_width, length, width, last_highlight = 0, highlight = 0;
 
   /* Find size of longest descriptor entry */
   length = width = 0;
@@ -306,33 +332,46 @@ int do_popup_menu(lde_menu menu[], lde_keymap keys[])
 
   while ( (c = mgetch()) ) {
     switch (c) {
-      case CTRL('P'):
+      case LDE_CTRL('P'):
       case KEY_UP:
       case 'k':
       case 'K':
 	if (highlight>0) highlight--;
 	break;
-      case CTRL('N'):
+      case LDE_CTRL('N'):
       case KEY_DOWN:
       case 'j':
       case 'J':
 	if (highlight<(window_length-1)) highlight++;
 	break;
-      case CTRL('A'):
-      case ESC:
+      case LDE_CTRL('A'):
+      case LDE_ESC:
 	delwin(win);
 	delwin(bigwin);
 	refresh_all();
 	return 0;
       case KEY_ENTER:
-      case CTRL('M'):
-      case CTRL('J'):
+      case LDE_CTRL('M'):
+      case LDE_CTRL('J'):
 	delwin(win);
 	delwin(bigwin);
 	refresh_all();
 	return menu[highlight].action_code;
       default:
-	if ( (c = lookup_key(c, keys)) != CMD_NO_ACTION ) {
+	if ( IS_LDE_MOUSE(c) ) {
+	  /* check x */
+	  result = 0;
+	  if ( LDE_MOUSEX(c) < window_width ) {
+	    c = LDE_MOUSEY(c) - HEADER_SIZE - 2;
+	    if ((c>=0)&&(c<length)) {
+	      result = menu[c].action_code; 
+	    }
+	  }
+	  delwin(win);
+	  delwin(bigwin);
+	  refresh_all();
+	  return result;
+	} else if ( (c = lookup_key(c, keys)) != CMD_NO_ACTION ) {
 	  delwin(win);
 	  delwin(bigwin);
 	  refresh_all();
@@ -592,7 +631,8 @@ void flag_popup(void)
 {
   WINDOW *win;
   int c, redraw, flag;
-  int vstart,vsize=8;
+  int vstart,vsize=9;
+  char *choices = "afnwi q";
 
   /* Do some bounds checking on our window */
   if (VERT>vsize) {
@@ -609,6 +649,16 @@ void flag_popup(void)
   while ((flag)||(c = mgetch())) {
     flag = 0;
     redraw = 0;
+
+    /* Allow mouse clicks */
+    if (IS_LDE_MOUSE(c)) {
+      if ( (LDE_MOUSEX(c)>HOFF) && (LDE_MOUSEY(c)>vstart) && (LDE_MOUSEX(c)<(HOFF+WIN_COL)) && (LDE_MOUSEY(c)<(vstart+vsize)) )  {
+	c=choices[LDE_MOUSEY(c)-vstart-2];
+      } else {
+	c = 'q';
+      }
+    }
+       
     switch (c) {
       case 'q':
       case 'Q':
@@ -639,7 +689,12 @@ void flag_popup(void)
         lde_flags.quiet = 1 - lde_flags.quiet;
 	redraw = 1;
         break;
-      case CTRL('L'):
+      case 'I':
+      case 'i':
+        lde_flags.blanked_indirects = 1 - lde_flags.blanked_indirects;
+	redraw = 1;
+        break;
+      case LDE_CTRL('L'):
         refresh_all();
         break;
       case ' ':
@@ -652,27 +707,32 @@ void flag_popup(void)
     box(win,0,0);
     wattroff(win,RED_ON_BLACK);
     mvwprintw(win,1,15,"A: (%-3s) Search all blocks",lde_flags.search_all ? "YES" : "NO");
-    mvwprintw(win,2,15,"F: (%-3s) Always append existing file ?",lde_flags.always_append ? "YES" : "NO");
+    mvwprintw(win,2,15,"F: (%-3s) Always append existing file?",lde_flags.always_append ? "YES" : "NO");
     mvwprintw(win,3,15,"N: (%-3s) Noise is off -- i.e. quiet",lde_flags.quiet ? "YES" : "NO");
     mvwprintw(win,4,15,"W: (%-3s) OK to write to file system",lde_flags.write_ok ? "YES" : "NO");
-    mvwprintw(win,6,15,"Q: return to editing");
+    mvwprintw(win,5,15,"I: (%-3s) Ignore indirect block contents (Linux <2.0.33 fix)",lde_flags.blanked_indirects ? "YES" : "NO");
+    mvwprintw(win,vsize-2,15,"Q: return to editing");
     wrefresh(win);
   }
 }
 
 /* Query the user for a file name, then ask for confirmation,
  * then dump the file from the list of inodes */
-void crecover_file(unsigned long inode_zones[])
+void crecover_file(unsigned long inode_zones[], unsigned long filesize)
 {
+  static char recover_file_name[80] = "RECOVERED.file";
   int fp;
-  static char recover_file_name[80];
   char recover_query[80];
-
-  if (recover_file_name[0] == 0) strcpy(recover_file_name,"RECOVERED.file");
+  int c;
 
   sprintf(recover_query," Write data to file:");
-  if (cread_num(recover_query, NULL))
+  c = cread_num(recover_query, NULL);
+  if (c<0) {
+    lde_warn("Recovery aborted");
+    return;
+  }  else if (c>0) {
     strncpy(recover_file_name, recover_query, 80);
+  }
 
   if ( (fp = open(recover_file_name,O_RDONLY)) > 0 ) {
     close(fp);
@@ -680,7 +740,7 @@ void crecover_file(unsigned long inode_zones[])
     if (lde_flags.always_append) {
       fp = open(recover_file_name,O_WRONLY|O_APPEND);
     } else {
-      switch (cquery("File exists, append data [Yes/Overwrite/Quit]: ","ynq","")) {
+      switch (cquery("File exists, append data [Yes/Overwrite/Quit]: ","yoq","")) {
         case 'y':
 	  fp = open(recover_file_name,O_WRONLY|O_APPEND);
 	  break;
@@ -696,9 +756,10 @@ void crecover_file(unsigned long inode_zones[])
     lde_warn("Cannot open file '%s'",recover_file_name);
 
   if (fp > 0) {
-    recover_file(fp, inode_zones);
+    lde_warn("Recovery in progress . . .");
+    if (!recover_file(fp, inode_zones, filesize))
+      lde_warn("Recovered data written to '%s'", recover_file_name);
     close(fp);
-    lde_warn("Recovered data written to '%s'", recover_file_name);
   } else {
     lde_warn("Error opening '%s'",recover_file_name);
   }
@@ -710,12 +771,13 @@ int recover_mode(void)
 {
   int j,c,next_cmd=CMD_REFRESH;
   unsigned long a;
-  char recover_labels[INODE_BLKS];
-  int vstart,vsize=fsc->N_BLOCKS;
+  char recover_labels[INODE_BLKS+1];
+  int vstart,vsize=fsc->N_BLOCKS+1;
 
   /* Fill in keys used to change blocks in recover mode */
   for (j=REC_FILE0; j<REC_FILE_LAST; j++)
     recover_labels[j-REC_FILE0] = (text_key(j,recover_keymap,0))[1];
+  recover_labels[INODE_BLKS] = (text_key(REC_FILE_SIZE,recover_keymap,0))[1];
 
   /* Do some bounds checking on our window */
   if (VERT>vsize) {
@@ -753,8 +815,12 @@ int recover_mode(void)
 	if (cread_num("Enter block number (leading 0x or $ indicates hex):", &a))
 	  fake_inode_zones[c-REC_FILE0] = (unsigned long) a;
         break;
+      case REC_FILE_SIZE:
+	if (cread_num("Enter file size (0 is special) (leading 0x or $ indicates hex):", &a))
+	  fake_inode_zones[INODE_BLKS] = (unsigned long) a;
+        break;
       case CMD_CLR_RECOVER:
-	bzero(fake_inode_zones,sizeof(long)*MAX_BLOCK_POINTER);
+	bzero(fake_inode_zones,sizeof(long)*(INODE_BLKS+1));
 	break;
       case CMD_BLOCK_MODE:
       case CMD_VIEW_SUPER:
@@ -775,10 +841,10 @@ int recover_mode(void)
 	refresh_all();
 	break;
       case CMD_DO_RECOVER:
-	crecover_file(fake_inode_zones);
+	crecover_file(fake_inode_zones, fake_inode_zones[INODE_BLKS]);
 	break;
       case CMD_CHECK_RECOVER:
-	(void) check_recover_file(fake_inode_zones);
+	(void) check_recover_file(fake_inode_zones, fake_inode_zones[INODE_BLKS]);
 	break;
       case CMD_DISPLAY_LOG:
 	error_popup();
@@ -805,6 +871,8 @@ int recover_mode(void)
       mvwprintw(workspace,j,40," %1c : 0x%8.8lX",recover_labels[j],fake_inode_zones[fsc->X3_INDIRECT]);
       j++;
     }
+    mvwprintw(workspace,j,20,"FILE SIZE:" );
+    mvwprintw(workspace,j,40," %1c : 0x%8.8lX",recover_labels[INODE_BLKS],fake_inode_zones[INODE_BLKS]);
     wrefresh(workspace);
   }	
 
@@ -849,6 +917,15 @@ int lookup_key(int c, lde_keymap *kmap)
 {
   int i;
 
+  /* A mouse click in the header window should popup a menu */
+  if (IS_LDE_MOUSE(c)) {
+    if (LDE_MOUSEY(c)<=HEADER_SIZE) {
+      return CMD_CALL_MENU;
+    } else {
+      return c&0xFFFFFF00;
+    }
+  }
+
   for (i=0;kmap[i].key_code;i++) {
     if (c==kmap[i].key_code) {
       return kmap[i].action_code;
@@ -902,12 +979,12 @@ char *text_key(int c, lde_keymap *kmap, int skip)
       memset(return_string,0,5);
 
       /* Look for meta and ctrl keys */
-      if (IS_META(c)) {
+      if (IS_LDE_META(c)) {
 	strcpy(return_string,"M-");
-	c = INV_META(c);
-      } else if (IS_CTRL(c)) {
+	c = INV_LDE_META(c);
+      } else if (IS_LDE_CTRL(c)) {
 	strcpy(return_string,"C-");
-	c = INV_CTRL(c);
+	c = INV_LDE_CTRL(c);
       } else {
 	strcpy(return_string," ");
       }
@@ -930,17 +1007,28 @@ char *text_key(int c, lde_keymap *kmap, int skip)
 /* Not too exciting main parser with superblock on screen */
 void interactive_main(void)
 {
+#ifdef HAVE_LIBGPM
+  Gpm_Connect gpmconn;
+#endif
+  SCREEN *ldeScreen;
   int c, next_cmd = CMD_DISPLAY_LOG;
 
+
   current_inode = fsc->ROOT_INODE;
   current_block = 0UL;
 
   /* Curses initialization junk */
-  if (!(newterm(NULL,stdout,stdin))) {
-    printf("It seems you have not set up ncurses correctly -- See INSTALL for assistance.\n");
+  if (!(ldeScreen=newterm(NULL,stdout,stdin))) {
+    printf("* It seems you have not set up ncurses correctly -- See INSTALL for assistance.\n");
+    if (termname()) {
+      printf("* Unknown terminal type '%s' specified.\n", termname());
+      printf("* Check your shell's TERM and TERMINFO variables.\n");
+    } else {
+      printf("* Your shell's TERM variable is unset.\n");
+    }
+    printf("* Try something like 'export TERM=vt100' or 'setenv TERM vt100'\n");
     return;
   }
-  initscr();
   cbreak();
   noecho();
   keypad(stdscr, TRUE);
@@ -962,24 +1050,43 @@ void interactive_main(void)
   /* First check to see that our screen is big enough */
   if ((LINES<(4+HEADER_SIZE+TRAILER_SIZE))||(COLS<WIN_COL)) {
     endwin();
+    delscreen(ldeScreen);
     printf("Screen is too small: need %d x %d, have %d x %d\n",
 	   WIN_COL,(4+HEADER_SIZE+TRAILER_SIZE),
 	   COLS,LINES);
     return;
   }
 
+  /* Now that ncurses has successfully initted, 
+     switch to curses based warning and getch functions */
+  lde_warn = nc_warn;
+  mgetch = nc_mgetch;
+
+# ifdef HAVE_LIBGPM  
+    /* Setup mouse handler */
+    gpmconn.eventMask=GPM_UP;
+    gpmconn.defaultMask=GPM_MOVE;
+    gpmconn.maxMod=0;
+    gpmconn.minMod=0;
+    Gpm_Open(&gpmconn,0);
+    gpm_handler = lde_mouse_handler;
+# endif
+
   /* Clear out restore buffer */
-  bzero(fake_inode_zones,sizeof(long)*MAX_BLOCK_POINTER);
+  bzero(fake_inode_zones,sizeof(long)*(INODE_BLKS+1));
  
   /* Our three curses windows */
-#if HEADER_SIZE>0
-  header = newwin(HEADER_SIZE,COLS,0,0);
-#endif
+# if HEADER_SIZE>0
+    header = newwin(HEADER_SIZE,COLS,0,0);
+    wbkgdset(header,WHITE_ON_BLUE);
+# endif
   workspace = newwin(1,0,0,0); /* Gets clobbered before it gets used */
-#if TRAILER_SIZE>0
-  trailer = newwin(TRAILER_SIZE,COLS,LINES-TRAILER_SIZE,0);
-#endif
+# if TRAILER_SIZE>0
+    trailer = newwin(TRAILER_SIZE,COLS,LINES-TRAILER_SIZE,0);
+    wbkgdset(trailer,WHITE_ON_BLUE);
+# endif
   
+  refresh_all();       /* NCurses 4.2 needs this refresh for some reason? */
   restore_header();
   
   while ( (c=next_cmd)||(c=lookup_key(mgetch(),main_keymap)) ) {
@@ -997,8 +1104,11 @@ void interactive_main(void)
 	continue;
       case CMD_EXIT_PROG:
 	endwin();
+	delscreen(ldeScreen);
+#       ifdef HAVE_LIBGPM
+	   Gpm_Close();
+#       endif
 	return;
-	break;
       case CMD_RECOVERY_MODE:
 	next_cmd = recover_mode();
 	continue;
