@@ -3,7 +3,7 @@
  *
  *  Copyright (C) 1994  Scott D. Heavner
  *
- *  $Id: recover.c,v 1.17 1996/10/18 12:17:05 sdh Exp $
+ *  $Id: recover.c,v 2.5 2000/03/11 19:07:59 denis Exp $
  */
 
 #include <stdio.h>
@@ -11,16 +11,19 @@
 #include <unistd.h>
 #include <ctype.h>
 
+#ifdef   HAVE_UNAME
+#include <sys/utsname.h>
+#endif
 #include <linux/fs.h>
 
 #include "lde.h"
 #include "recover.h"
 #include "tty_lde.h"
 
-/* This takes care of the mapping from a char pointer to unsigned long/short, depending
- * on the file system.
- */
-unsigned long block_pointer(unsigned char *ind, unsigned long blknr, int zone_entry_size)
+/* This takes care of the mapping from a char pointer to unsigned
+ * long/short, depending on the file system */
+unsigned long block_pointer(unsigned char *ind, 
+			    unsigned long blknr, int zone_entry_size)
 {
   unsigned long  *lind;
   unsigned short *sind;
@@ -34,10 +37,115 @@ unsigned long block_pointer(unsigned char *ind, unsigned long blknr, int zone_en
   }
 }
 
+#ifdef BETA_CODE
+/* Try to work around Linux <= 2.0.33 bug */ 
 /* This is used to pull the correct block from the inode block table which
- * should have been copied into zone_index. 
- */
-int map_block(unsigned long zone_index[], unsigned long blknr, unsigned long *mapped_block)
+ * should have been copied into zone_index */
+static int hacked_map_block_helper(unsigned long zone_index[],
+				   unsigned long blknr,
+				   unsigned long *mapped_block)
+{
+  long last_direct = 0, i, offset = 0;
+
+  *mapped_block = 0UL;
+  offset = blknr;
+
+  if (fsc->N_DIRECT) {
+    /* Direct blocks: Same as map_block() */
+    if (blknr<fsc->N_DIRECT) {
+      *mapped_block = zone_index[blknr];
+      if ( zone_index[blknr] < sb->nzones ) {
+	return (EMB_NO_ERROR);
+      } else {
+	return (EMB_DIRECT_RANGE);
+      }
+    }
+    offset -= fsc->N_DIRECT;
+    blknr  -= (fsc->N_DIRECT-1); /* This is the only place I subtract from 
+				  * blknr: want to reference from last used
+				  * direct block */
+  } else {
+    /* If no direct block we're hosed */
+    return (EMB_WAY_OUT_OF_RANGE);
+  }    
+  
+  /* Need to find last used direct block */
+  for (i=0; i<fsc->N_DIRECT; i++)
+    if (zone_index[i])
+      last_direct = zone_index[i];
+    
+  /* If no used directs, we're really screwed */
+  if (last_direct) {
+    if (fsc->INDIRECT) {
+      blknr++;       /* Skip data for indirect block */   
+      if (offset<ZONES_PER_BLOCK) {
+	/* Skip indirect block contents */
+	*mapped_block = last_direct + blknr;
+	return EMB_NO_ERROR;
+      }
+      offset -= ZONES_PER_BLOCK;
+    }
+    if (fsc->X2_INDIRECT) {
+      blknr++;       /* Skip data for 2x indirect block */
+      if (offset<(ZONES_PER_BLOCK*ZONES_PER_BLOCK)) {
+	/* Already skipped 2x, but need to adjust for any 1x's we're past, 
+	 * offset starts at 0, so always add 1 to offset/ZONES_PER_BLOCK */
+	*mapped_block = last_direct + blknr + (offset/ZONES_PER_BLOCK+1);
+	return EMB_NO_ERROR;
+      }
+      offset -= ZONES_PER_BLOCK*ZONES_PER_BLOCK;
+      blknr  += ZONES_PER_BLOCK;   /* There were ZONES_PER_BLOCK indirect index blocks we skipped over */ 
+    }
+    if (fsc->X3_INDIRECT) {
+      blknr++;       /* Skip data for 3x indirect block */
+      if (offset<(ZONES_PER_BLOCK*ZONES_PER_BLOCK*ZONES_PER_BLOCK)) {
+	/* Need to multiple 2x indirect and any 1x's we're passed */
+	*mapped_block = last_direct + blknr +
+	  (offset/ZONES_PER_BLOCK/ZONES_PER_BLOCK+1) + /* num 2x's */
+	  (offset/ZONES_PER_BLOCK+1);                 /* num 1x's */
+	return EMB_NO_ERROR;
+      }
+      offset -= ZONES_PER_BLOCK*ZONES_PER_BLOCK*ZONES_PER_BLOCK; 
+      blknr  += ZONES_PER_BLOCK*ZONES_PER_BLOCK + ZONES_PER_BLOCK; 
+    }
+  }
+
+  /* Nothing */
+  return (EMB_WAY_OUT_OF_RANGE);
+}
+
+static int hacked_map_block(unsigned long zone_index[],
+			    unsigned long blknr,
+			    unsigned long *mapped_block, 
+			    unsigned long *skipped_block)
+{
+  int result;
+
+  /* Lookup block, return any error */
+  result = hacked_map_block_helper(zone_index,blknr,mapped_block);
+  if (result!=EMB_NO_ERROR)
+    return result;
+
+  *mapped_block += *skipped_block;
+
+  /* Check for system/used blocks and skip 
+   * (system blocks will be marked used, duh!
+   *  FS_cmd.is_system_block seems like a waste) */
+  while ( ((!lde_flags.search_all)&&FS_cmd.zone_in_use(*mapped_block)) ||
+	  FS_cmd.is_system_block(*mapped_block) ) {
+    (*skipped_block)++;
+    if ( ++(*mapped_block) > sb->nzones)
+      return (EMB_WAY_OUT_OF_RANGE);
+  }
+
+  return EMB_NO_ERROR;
+}
+#endif
+
+/* This is used to pull the correct block from the inode block table which
+ * should have been copied into zone_index */
+int map_block(unsigned long zone_index[], unsigned long blknr,
+	      unsigned long *mapped_block)
 {
   unsigned char *ind = NULL;
   unsigned long block;
@@ -57,7 +165,6 @@ int map_block(unsigned long zone_index[], unsigned long blknr, unsigned long *ma
     blknr -= fsc->N_DIRECT;
   }
 
-
   /* Ok, it wasn't a direct block.  Move on to indirect block(s) */
   if (fsc->INDIRECT) {
     if (blknr<ZONES_PER_BLOCK) {
@@ -68,7 +175,7 @@ int map_block(unsigned long zone_index[], unsigned long blknr, unsigned long *ma
 	return (EMB_IND_ZERO);
       } else if (block<sb->nzones) {
 	/* Indirect block is in fs range, now read it in */
-	ind = cache_read_block(block,CACHEABLE);
+	ind = cache_read_block(block,NULL,CACHEABLE);
 	*mapped_block = block_pointer(ind,blknr,fsc->ZONE_ENTRY_SIZE);
 	if (*mapped_block >= sb->nzones) {
 	  return (EMB_IND_LOOKED_RANGE);
@@ -91,16 +198,19 @@ int map_block(unsigned long zone_index[], unsigned long blknr, unsigned long *ma
 	*mapped_block = fsc->N_DIRECT + ZONES_PER_BLOCK*(ZONES_PER_BLOCK+1);
 	return (EMB_2IND_ZERO);
       } else if (block<sb->nzones) {
-	ind = cache_read_block(block,CACHEABLE);
-	block = block_pointer(ind,(unsigned long)(blknr/ZONES_PER_BLOCK),fsc->ZONE_ENTRY_SIZE);
+	ind = cache_read_block(block,NULL,CACHEABLE);
+	block = block_pointer(ind,(unsigned long)(blknr/ZONES_PER_BLOCK),
+			      fsc->ZONE_ENTRY_SIZE);
 	if (block==0UL) {
                                                          /* 2 here, 1 for indirect block, 1 for current 2xindirect */
 	  *mapped_block = fsc->N_DIRECT + ZONES_PER_BLOCK*(2+blknr/ZONES_PER_BLOCK); 
 	  /* *mapped_block = fsc->N_DIRECT + ZONES_PER_BLOCK*2 + blknr; */
 	  return (EMB_2IND_L1_ZERO);
 	} else if (block<sb->nzones) {
-	  ind = cache_read_block(block,CACHEABLE);
-	  *mapped_block = block_pointer(ind,(unsigned long)(blknr%ZONES_PER_BLOCK),fsc->ZONE_ENTRY_SIZE);
+	  ind = cache_read_block(block,NULL,CACHEABLE);
+	  *mapped_block = 
+	    block_pointer(ind,(unsigned long)
+			  (blknr%ZONES_PER_BLOCK),fsc->ZONE_ENTRY_SIZE);
 	  if (*mapped_block >= sb->nzones) {
 	    *mapped_block = 0UL;
 	    return (EMB_2IND_LOOKED_RANGE);
@@ -108,7 +218,8 @@ int map_block(unsigned long zone_index[], unsigned long blknr, unsigned long *ma
 	    return (EMB_NO_ERROR);
 	  }
 	} else {
-	  *mapped_block = fsc->N_DIRECT + ZONES_PER_BLOCK*(2+blknr/ZONES_PER_BLOCK);
+	  *mapped_block = fsc->N_DIRECT +
+	    ZONES_PER_BLOCK*(2+blknr/ZONES_PER_BLOCK);
 	  /* *mapped_block = fsc->N_DIRECT + ZONES_PER_BLOCK*2 + blknr; */
 	  return (EMB_2IND_L1_RANGE);
 	}
@@ -120,8 +231,8 @@ int map_block(unsigned long zone_index[], unsigned long blknr, unsigned long *ma
     blknr -= (ZONES_PER_BLOCK*ZONES_PER_BLOCK);
   }
 
-  /* Shouldn't we be able to do this recursively??  It can do indirect and 2x indirects, whats another
-   * level of indirection? easy.
+  /* Shouldn't we be able to do this recursively??  It can do indirect 
+   * and 2x indirects, whats another level of indirection? easy.
    */
   if (fsc->X3_INDIRECT) {
     if (blknr<(ZONES_PER_BLOCK*ZONES_PER_BLOCK*ZONES_PER_BLOCK)) {
@@ -133,18 +244,20 @@ int map_block(unsigned long zone_index[], unsigned long blknr, unsigned long *ma
 #ifdef BETA_CODE
       } else if (block<sb->nzones) {
 	/* Read in the triple indirect block */
-	ind = cache_read_block(block,CACHEABLE);
-	block = block_pointer(ind,(unsigned long)(blknr/(ZONES_PER_BLOCK*ZONES_PER_BLOCK)),
+	ind = cache_read_block(block,NULL,CACHEABLE);
+	block = block_pointer(ind,(unsigned long)
+			      (blknr/(ZONES_PER_BLOCK*ZONES_PER_BLOCK)),
 			      fsc->ZONE_ENTRY_SIZE);
 	/* Block is now a pointer to a 2x indirect */
 	if (block==0UL) {
 	  /* Add extra 2x indirect block to *mapped block */
 	  *mapped_block = fsc->N_DIRECT + ZONES_PER_BLOCK +  
-	    ZONES_PER_BLOCK*ZONES_PER_BLOCK*(2 + blknr/(ZONES_PER_BLOCK*ZONES_PER_BLOCK));
+	    ZONES_PER_BLOCK*
+	    ZONES_PER_BLOCK*(2 + blknr/(ZONES_PER_BLOCK*ZONES_PER_BLOCK));
 	  return (EMB_3IND_L1_ZERO);
 	} else if (block<sb->nzones) {
 	  /* Read in a 2x indirect */
-	  ind = cache_read_block(block,CACHEABLE);
+	  ind = cache_read_block(block,NULL,CACHEABLE);
 	  block = block_pointer(ind,  (unsigned long)
 				((blknr%(ZONES_PER_BLOCK*ZONES_PER_BLOCK))/ZONES_PER_BLOCK),
 				fsc->ZONE_ENTRY_SIZE);
@@ -156,7 +269,7 @@ int map_block(unsigned long zone_index[], unsigned long blknr, unsigned long *ma
 	    return (EMB_3IND_L2_ZERO);
 	  } else if (block<sb->nzones) {
 	    /* Read in 1x indirect */
-	    ind = cache_read_block(block,CACHEABLE);
+	    ind = cache_read_block(block,NULL,CACHEABLE);
 	    *mapped_block = block_pointer(ind,(unsigned long)
 					  ((blknr%(ZONES_PER_BLOCK*ZONES_PER_BLOCK))%ZONES_PER_BLOCK),
 					  fsc->ZONE_ENTRY_SIZE);
@@ -205,21 +318,46 @@ int advance_zone_pointer(unsigned long zone_index[], unsigned long *currblk,
 		     unsigned long *ipointer, long increment)
 {
   int result;
-  unsigned long blknr = 0UL, local_ipointer;
+  unsigned long blknr = 0UL, local_ipointer=0UL;
+  long sincrement,i;
 
   /* Check that we are at a block that is indexed by this inode */
-  if ((map_block(zone_index, *ipointer, &blknr))||(blknr!=(*currblk)))
-    return AZP_BAD_START;
+  if ((map_block(zone_index, *ipointer, &blknr))||(blknr!=(*currblk))) {
+    /* We aren't: look it up using brute force */
+    for (;;) {
+      result = map_block(zone_index, local_ipointer, &blknr);
+      if (result < EMB_HALT) {
+	return AZP_BAD_START;
+      } else if (result < EMB_SKIP) {
+	local_ipointer = blknr;
+      } else if ((result==EMB_NO_ERROR) && (blknr==(*currblk))) {
+	break;
+      } else {
+	local_ipointer++;
+      }
+    }
+  } else {
+    local_ipointer = *ipointer;
+  }
 
-  local_ipointer = *ipointer;
+  sincrement = (increment < 0)?-1:1;
+  increment *= sincrement;
 
   /* Now adjust the pointer and find us a block */
-  do {
-    local_ipointer += increment;
-    result = map_block(zone_index, local_ipointer, &blknr);
-    if (result < EMB_HALT)
-      return result;
-  } while(result!=EMB_NO_ERROR);
+  for (i=0; i<increment; i++) {
+    for (;;) {
+      local_ipointer += sincrement;
+      result = map_block(zone_index, local_ipointer, &blknr);
+      if (result < EMB_HALT) {
+	return result;
+      } else if (result < EMB_SKIP) {
+	local_ipointer = blknr - sincrement;
+	continue;
+      } else if (result==EMB_NO_ERROR) {
+	break;
+      }
+    }
+  }
 
   if (blknr) {
     *ipointer = local_ipointer;
@@ -240,59 +378,111 @@ int advance_zone_pointer(unsigned long zone_index[], unsigned long *currblk,
  * map is zero, I will dump block zero, in any other position,
  * it signals the end of the map.
  */
-void recover_file(int fp,unsigned long zone_index[])
+int recover_file(int fp,unsigned long zone_index[],unsigned long filesize)
 {
   unsigned char *dind;
-  unsigned long nr;
+  unsigned long nr, written=0UL, skipped=0UL;
   int j, result;
   size_t write_count;
 
   j = 0;
   lde_flags.quit_now = 0;
+
+  /* Lookup system version, warn about linux 2.0.33 */
+#ifdef HAVE_UNAME /* Why bother with a HAVE_UNAME check,
+                   *  I don't check for Linux name, just version */
+  if (!lde_flags.blanked_indirects) {
+    struct utsname nm;
+    uname(&nm);
+    if ( (nm.release[0]=='2')&&(nm.release[2]=='0') ) {
+      lde_warn("Linux 2.0.* users should activate the blanked_indirects"
+	       " flag for proper recovery!  See the manual!");
+      sleep(5);
+    }
+  }
+#endif
+
   while (1) {
-    if (lde_flags.quit_now)
-      break;
-    if ((result=map_block(zone_index,j,&nr))) {
+
+    /* User has pressed ctrl-c, abort recover */
+    if (lde_flags.quit_now) {
+      lde_warn("User aborted recovery, partial data may have been written to file.");
+      return -2;
+    }
+
+    /* Lookup block number from inode */
+#ifdef BETA_CODE
+    if (lde_flags.blanked_indirects) 
+      result = hacked_map_block(zone_index,j,&nr,&skipped);
+    else 
+      result=map_block(zone_index,j,&nr);
+#else
+    result=map_block(zone_index,j,&nr);
+#endif
+
+    /* Block successfully looked up? */
+    if (result) {
       if (result < EMB_HALT) {
-	break;
+	return 0;  /* we hit last entry in this inode, return success */
       } else if (result < EMB_SKIP ) {
-	j = nr;
+	j = nr;    /* block has zero entry, go back and get another one */
 	continue;
       } else {
-	nr = 0UL;
+	nr = 0UL;  /* Don't do any processing on this block, but fall 
+		    * through to increment j */
       }
     }
 
-    dind = cache_read_block(nr,CACHEABLE);
+    /* Have a valid block number? */
     if (nr) {
+      /* Read data from disk */
+      dind = cache_read_block(nr,NULL,CACHEABLE);
+      /* Check for small block if we opened a file instead of a device */
       write_count = (size_t) lookup_blocksize(nr);
-      /* lde_warn("Setting write count block size %d",write_count); */
+      /* Add what we're about to write to total written */
+      written += write_count;
+      /* If gone to far, don't want to write out all this data */
+      if ((filesize)&&(written>filesize))
+	write_count -= written - filesize;
+      /* Write out data to recovery file */
       if (write (fp, dind, write_count) != write_count) {
 	lde_warn("Write error: unable to write block (%ld) to recover file, recover aborted",nr);
-	return;
+	return -1;
       }
+      /* Quit when we've written enough */
+      if ((filesize)&&(written>=filesize))
+	return 0;
     }
     j++;
   }
+
+  return 0;
 }
 
 /* Goes through and checks to see if the file can be recovered.  I.e. is another
  * file using blocks that were once used by this inode */
-int check_recover_file(unsigned long zone_index[])
+int check_recover_file(unsigned long zone_index[],unsigned long filesize)
 {
-  unsigned long nr, lookedup=0UL;
+  unsigned long nr, lookedup=0UL, checked=0UL;
   int j, result, cr_result = 0;
-  
   j = 0;
   lde_flags.quit_now = 0;
 
   while (1) {
+    /* Stop if we specified a filesize and have covered that many blocks */
+    if ((filesize)&&(checked>filesize))
+      break;
+    
+    /* Terminate search on ^C */
     if (lde_flags.quit_now) {
       lde_flags.quit_now = 0;
       lde_warn("Search terminated.");
       return 1;
     }
-    if ((result=map_block(zone_index,j,&nr))) {
+
+    result=map_block(zone_index,j,&nr);
+
+    if (result) {
       if (result < EMB_HALT) {
 	break;
       } else if (result < EMB_SKIP ) {
@@ -301,7 +491,10 @@ int check_recover_file(unsigned long zone_index[])
       } else {
 	nr = 0UL;
       }
+    } else {
+      checked += lookup_blocksize(nr);
     }
+
     if ((nr)&&FS_cmd.zone_in_use(nr)) {
       lde_warn("Block %ld (0x%lX) in use by another file. Hit any key to continue (q=abort, l=lookup inode).",nr,nr);
       result = mgetch();
@@ -427,7 +620,7 @@ void search_fs(unsigned char *search_string, int search_len, int search_off, uns
   lde_flags.quit_now = 0;
 
   /* Where should we start the search? */
-  if (start_nr!=1UL) {
+  if (start_nr!=0UL) {
     lde_warn("Resuming search from block 0x%lX",start_nr);
     nr = start_nr;
   } else {
@@ -443,7 +636,7 @@ void search_fs(unsigned char *search_string, int search_len, int search_off, uns
 
     /* Do all the searches in the unused data space */
     if ((!FS_cmd.zone_in_use(nr))||(lde_flags.search_all)) {
-      dind = cache_read_block(nr,CACHEABLE);
+      dind = cache_read_block(nr,NULL,CACHEABLE);
 
       /* Search codes --
        * see /etc/magic or make a similar file 
@@ -464,18 +657,19 @@ void search_fs(unsigned char *search_string, int search_len, int search_off, uns
 	
 	if (matched) {
 	  printf("Match at block 0x%lX",nr);
-	  if (lde_flags.inode_lookup)
+	  if (lde_flags.inode_lookup) {
 	    if ( (inode_nr = find_inode(nr, 0UL)) ) {
 	      printf(", check inode 0x%lX",inode_nr);
 	      if (lde_flags.check_recover) {
 		lde_warn = no_warn;  /* Suppress output */
 		GInode = FS_cmd.read_inode(nr);
-		printf(", recovery %spossible",(check_recover_file(GInode->i_zone)?"":"NOT ") );
+		printf(", recovery %spossible",(check_recover_file(GInode->i_zone,GInode->i_size)?"":"NOT ") );
 		lde_warn = tty_warn; /* Reinstate output */
 	      }
 	    } else {
 	      printf(", no %sinode found",((lde_flags.search_all)?"":"unused ") );
 	    }
+	  }
 	  printf(".\n");
 	}
       }
